import * as gtv from "../gtv";
import { signDigest, checkDigestSignature } from "../encryption/encryption";
import { AlreadySignedTransactionException, MissingGtxException, MissingSignerException, NumberOfSignersAndSignaturesException, } from "./errors";
import { encode as gtxEncode, decode as gtxDecode } from "./index";
export function emptyGtx(blockchainRID) {
    return { blockchainRID: blockchainRID, operations: [], signers: [] };
}
/**
 * Adds a function call to a GTX. Creates a new GTX if none specified.
 * This function will throw Error if gtx is already signed
 * @param opName the name of the function to call
 * @param args the array of arguments of the function call. If no args, this must be an empty array
 * @param gtx the function call will be added to this gtx
 * @returns the gtx
 * @throws if gtx is null or if gtx is already signed
 */
export function addTransactionToGtx(opName, args, gtx) {
    if (gtx == null) {
        throw new MissingGtxException();
    }
    if (gtx.signatures) {
        throw new AlreadySignedTransactionException("function");
    }
    gtx.operations.push({ opName: opName, args: args });
    return gtx;
}
export function addSignerToGtx(signer, gtx) {
    if (gtx.signatures) {
        throw new AlreadySignedTransactionException("signer");
    }
    gtx.signers.push(signer);
}
/**
 * Serializes the gtx for signing
 * @param gtx the gtx to serialize
 */
export function getDigestToSign(gtx) {
    return gtv.gtvHash(gtvTxBody(gtx));
}
export function gtvTxBody(gtx) {
    return [
        gtx.blockchainRID,
        gtx.operations.map((op) => [op.opName, op.args]),
        gtx.signers,
    ];
}
/**
 * Signs the gtx with the provided privKey. This is a convenience function
 * for situations where you don't have to ask someone else to sign.
 */
export function sign(privKey, pubKey, gtx) {
    const digestToSign = getDigestToSign(gtx);
    const signature = signDigest(digestToSign, privKey);
    addSignature(pubKey, signature, gtx);
}
export function signRawTransaction(_keyPair, _rawTransaction) {
    throw Error("TODO");
    //TODO
    //const gtx = module.exports.deserialize(rawTransaction);
    //module.exports.sign(keyPair.privKey, keyPair.pubKey, gtx);
    // return module.exports.serialize(gtx)
}
/**
 * Adds a signature to the gtx
 */
export function addSignature(pubKeyBuffer, signatureBuffer, gtx) {
    if (!gtx.signatures) {
        gtx.signatures = Array(gtx.signers.length).fill(Buffer.alloc(64));
    }
    if (gtx.signers.length !== gtx.signatures.length) {
        throw new NumberOfSignersAndSignaturesException();
    }
    const signerIndex = gtx.signers.findIndex((signer) => pubKeyBuffer.equals(signer));
    if (signerIndex === -1) {
        throw new MissingSignerException();
    }
    gtx.signatures[signerIndex] = signatureBuffer;
}
export function serialize(gtx) {
    if (!gtx.signatures) {
        // TODO
        // The gtx is not signed, but we must include
        // the signatures attribute, so let's add that.
        gtx.signatures = [];
    }
    return gtxEncode([gtvTxBody(gtx), gtx.signatures]);
}
export function deserialize(gtxBytes) {
    const deserializedTx = gtxDecode(gtxBytes);
    const body = deserializedTx[0];
    const gtvTxBody = {
        blockchainRID: body[0],
        operations: body[1].map((operation) => ({
            opName: operation[0],
            args: operation[1],
        })),
        signers: body[2],
    };
    const signatures = deserializedTx[1];
    return {
        blockchainRID: gtvTxBody.blockchainRID,
        operations: gtvTxBody.operations,
        signers: gtvTxBody.signers,
        signatures,
    };
}
export function checkGTXSignatures(txHash, gtx) {
    for (const i in gtx.signers) {
        const signValid = checkDigestSignature(txHash, gtx.signers[i], gtx.signatures[i]);
        if (!signValid)
            return signValid;
    }
    return true;
}
//# sourceMappingURL=gtx.js.map