"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { ResponseStatus, Method, } from "./types";
import * as logger from "../logger";
import { EmptyListOfUrlsException, LostMessageError, SerializedTransactionFormatException, TxRejectedError, UnexpectedResponseError, UnexpectedStatusError, UnvalidTxRidException, } from "./errors";
import { toBuffer, toQueryObjectGTV } from "../formatter";
import { requestWithRetry } from "./restclientutil";
import { encodeValue } from "../gtx/serialization";
export function createRestClient(endpointPool, blockchainRID, maxSockets = 10, pollingInterval = 500, failOverConfig) {
    validateInput(endpointPool, failOverConfig);
    return {
        config: {
            endpointPool: endpointPool,
            pool: { maxSockets },
            pollingInterval,
            attemptsPerEndpoint: (failOverConfig === null || failOverConfig === void 0 ? void 0 : failOverConfig.attemptsPerEndpoint) || 3,
            attemptInterval: (failOverConfig === null || failOverConfig === void 0 ? void 0 : failOverConfig.attemptInterval) || 500,
        },
        /**
         * Retrieves the client message with the specified double-sha256 hash
         * @param txRID the id of the transaction
         * @param callback parameters (error, serializedMessage) if first
         * parameter is not null, an error occurred.
         * If first parameter is null, then the second parameter is a buffer
         * with the serialized client message. If no such client message exists,
         * the callback will be called with (null, null).
         */
        getTransaction: function (txRID, callback) {
            return __awaiter(this, void 0, void 0, function* () {
                if (!validTxRID(txRID)) {
                    callback(new UnvalidTxRidException(txRID), null);
                }
                else {
                    const { error, statusCode, rspBody } = yield requestWithRetry(Method.GET, "tx/" + blockchainRID + "/" + txRID.toString("hex"), this.config);
                    handleGetResponse(error, statusCode, statusCode === 200 ? toBuffer(rspBody.tx) : rspBody, callback);
                }
            });
        },
        /**
         * Sends a transaction to postchain for inclusion in a block.
         * Use status() to monitor progress once this transaction is
         * posted.
         *
         * @param serializedTransaction The transaction (a buffer) to send
         * @param callback taking parameter (error, responseObject) if error is null
         * then resonseObject is also null. If error is not null, then responseObject
         * is an object with the string property 'error'
         */
        postTransaction: function (serializedTransaction, callback) {
            return __awaiter(this, void 0, void 0, function* () {
                if (!Buffer.isBuffer(serializedTransaction)) {
                    throw new SerializedTransactionFormatException();
                }
                const transactionObject = {
                    tx: serializedTransaction.toString("hex"),
                };
                const { error, statusCode, rspBody } = yield requestWithRetry(Method.POST, "tx/" + blockchainRID, this.config, transactionObject);
                handlePostResponse(error, statusCode, rspBody, callback);
            });
        },
        /**
         * Retrieves a confirmation proof for a client message with the specified double-sha256
         * hash.
         * @param txRID the id of the transaction
         * @param callback parameters (error, responseObjectProof) if first
         * parameter is not null, an error occurred.
         * If first parameter is null, then the second parameter is an object
         * like the following:
         *
         * {hash: messageHashBuffer,
         *  blockHeader: blockHeaderBuffer,
         *  signatures: [{pubKey: pubKeyBuffer, signature: sigBuffer}, ...],
         *  merklePath: [{side: <0|1>, hash: <hash buffer level n-1>},
         *               ...
         *               {side: <0|1>, hash: <hash buffer level 1>}]}
         *
         * If no such client message exists, the callback will be called with (null, null).
         *
         * The proof object can be validated using
         * postchain-common.util.validateMerklePath(proof.merklePath, proof.hash,
         * proof.blockHeader.slice(32, 64))
         *
         * The signatures must be validated agains some know trusted source for valid signers
         * at this specific block height.
         */
        getConfirmationProof: function (txRID, callback) {
            return __awaiter(this, void 0, void 0, function* () {
                if (!validTxRID(txRID)) {
                    callback(new UnvalidTxRidException(txRID), null);
                }
                else {
                    const { error, statusCode, rspBody } = yield requestWithRetry(Method.GET, "tx/" +
                        blockchainRID +
                        "/" +
                        txRID.toString("hex") +
                        "/confirmationProof", this.config);
                    const confirmationProof = {
                        hash: undefined,
                        blockHeader: undefined,
                        signatures: [],
                        merklePath: [],
                    };
                    if (statusCode === 200 || statusCode === 503) {
                        confirmationProof.hash = toBuffer(rspBody.hash);
                        confirmationProof.blockHeader = toBuffer(rspBody.blockHeader);
                        if (rspBody.signatures) {
                            confirmationProof.signatures = rspBody.signatures.map((item) => ({
                                pubKey: toBuffer(item.pubKey),
                                signature: toBuffer(item.signature),
                            }));
                        }
                        if (rspBody.merklePath) {
                            confirmationProof.merklePath = rspBody.merklePath.map((item) => {
                                return { side: item.side, hash: toBuffer(item.hash) };
                            });
                        }
                    }
                    handleGetResponse(error, statusCode, confirmationProof, callback);
                }
            });
        },
        /**
         * Queries the status of a certain transaction.
         * @param txRID the id of the transaction
         * @param callback taking parameters (error, responseBody). If error is null
         * then responseBody is an object on the form
         * { status: '<confirmed|waiting|rejected|unknown>' }
         * If error is not null, then responseBody
         * is an object with the string property 'error'
         */
        status: function (txRID, callback) {
            return __awaiter(this, void 0, void 0, function* () {
                if (!validTxRID(txRID)) {
                    callback(new UnvalidTxRidException(txRID), null);
                }
                else {
                    const { error, statusCode, rspBody } = yield requestWithRetry(Method.GET, "tx/" + blockchainRID + "/" + txRID.toString("hex") + "/status", this.config);
                    handleGetResponse(error, statusCode, rspBody, callback);
                }
            });
        },
        /**
         * Interfaces the query enpoint of the Rell backend. Returns either a resolved or rejected promise.
         * @param queryObject an object that must contain a "type" and follows this pattern: { type: "nameOfQuery", arg1: argValue1, arg2: argvalue2 }
         */
        query: function (queryObject) {
            return __awaiter(this, void 0, void 0, function* () {
                // eslint-disable-next-line no-async-promise-executor
                return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                    const callback = (error, result) => {
                        if (error) {
                            reject(error);
                        }
                        else {
                            resolve(result);
                        }
                    };
                    const { error, statusCode, rspBody } = yield requestWithRetry(Method.POST, `query_gtv/${blockchainRID}`, this.config, encodeValue(toQueryObjectGTV(queryObject)));
                    handlePostResponse(error, statusCode, rspBody, callback);
                }));
            });
        },
        /**
         * Polls for status while waiting for response; confirmed, rejected or unknown. Returns either a resolved or rejected promise.
         * @param txRID the id of the transaction
         */
        waitConfirmation(txRID) {
            return new Promise((resolve, reject) => {
                this.status(txRID, (error, result) => {
                    if (error) {
                        reject(error);
                    }
                    else {
                        const status = result.status;
                        switch (status) {
                            case ResponseStatus.Confirmed:
                                resolve(null);
                                break;
                            case ResponseStatus.Rejected:
                                reject(new TxRejectedError(result.rejectReason));
                                break;
                            case ResponseStatus.Unknown:
                                reject(new LostMessageError());
                                break;
                            case ResponseStatus.Waiting:
                                setTimeout(() => this.waitConfirmation(txRID).then(resolve, reject), this.config.pollingInterval);
                                break;
                            default:
                                logger.error(status);
                                reject(new UnexpectedResponseError());
                        }
                    }
                });
            });
        },
        /**
         * Posts a transaction and polls for status while waiting for status response; confirmed, rejected or unknown. Returns either a resolved or rejected promise.
         * @param serializedTransaction The transaction (a buffer) to send
         * @param txRID the id of the transaction
         * @param validate true if the transaction needs to be validated
         */
        postAndWaitConfirmation(serializedTransaction, txRID, validate) {
            if (validate === true) {
                return Promise.reject("Automatic validation is not yet implemented");
            }
            return new Promise((resolve, reject) => {
                this.postTransaction(serializedTransaction, (error) => {
                    if (error)
                        reject(error);
                    else {
                        setTimeout(() => this.waitConfirmation(txRID).then(resolve, reject), 1011);
                    }
                });
            });
        },
        /**
         * Returns a string array with the endpoints hosting the dApp the client
         * is connected to
         */
        getEndpointPool() {
            return this.config.endpointPool;
        },
    };
}
/**
 * Validates that txRID is a Buffer of 32 bytes.
 * @param txRID A buffer of 32 bytes
 */
function validTxRID(txRID) {
    if (txRID.length != 32) {
        const error = new UnvalidTxRidException(txRID);
        logger.error(error.toString());
        return false;
    }
    return true;
}
function validateInput(endpointPool, failOverConfig) {
    if (!endpointPool.length) {
        throw new EmptyListOfUrlsException();
    }
    if ((failOverConfig === null || failOverConfig === void 0 ? void 0 : failOverConfig.attemptsPerEndpoint) < 1) {
        logger.debug("Attempts can not be 0 or below, setting it to 1");
        failOverConfig.attemptsPerEndpoint = 1;
    }
}
/**
 * @param error response error
 * @param statusCode response status code
 * @param responseObject the responsebody from the server
 * @param callback the callback function to propagate the error and response back to the caller
 */
function handleGetResponse(error, statusCode, responseObject, callback) {
    try {
        logger.debug(`error: ${error}, status code: ${statusCode}, response body: ${JSON.stringify(responseObject)}`);
        if (error) {
            callback(error, null);
        }
        else if (statusCode === 404) {
            logger.error("404 received");
            callback(null, null);
        }
        else if (statusCode != 200) {
            callback(new UnexpectedStatusError(statusCode), responseObject);
        }
        else {
            callback(null, responseObject);
        }
    }
    catch (error) {
        logger.error("restclient.handleGetResponse(): Failed to call callback function " +
            error);
    }
}
/**
 * @param error response error
 * @param statusCode response status code
 * @param responseObject the responsebody from the server
 * @param callback the callback function to propagate the error and response back to the caller
 */
function handlePostResponse(error, statusCode, responseObject, callback) {
    logger.debug(`error: ${error}, status code: ${statusCode}, response body: ${JSON.stringify(responseObject)}`);
    try {
        if (error) {
            logger.error(`In restclient post(). ${error}`);
            callback(error);
        }
        else if (statusCode != 200) {
            logger.error(`Unexpected status code from server: ${statusCode}`);
            callback(new UnexpectedStatusError(statusCode), responseObject);
        }
        else {
            logger.info(`Calling responseCallback with responseObject: ${JSON.stringify(responseObject)}`);
            callback(null, responseObject);
        }
    }
    catch (error) {
        logger.error("restclient.handlePostResponse(): Failed to call callback function " +
            error);
    }
}
//# sourceMappingURL=restclient.js.map